import { Logger as WinstonLogger } from 'winston';
import { Logger as TypeOrmLogger } from 'typeorm';
import { requestContext } from '@app/common/context/request-context';

const ERROR_THRESHOLD = 2000;

export class TypeOrmWinstonLogger implements TypeOrmLogger {
  constructor(private readonly logger: WinstonLogger) {}

  private getContext() {
    const context = requestContext.getStore();
    return {
      requestId: context?.requestId ?? 'N/A',
      userId: context?.userId ?? null,
    };
  }

  private extractTable(query: string): string {
    const match =
      query.match(/from\s+["`]?(\w+)["`]?/i) ||
      query.match(/into\s+["`]?(\w+)["`]?/i);
    return match?.[1] ?? 'unknown';
  }

  logQuery(query: string, parameters?: any[]) {
    const context = this.getContext();
    this.logger.debug(`[DB][QUERY] ${query}`, {
      ...context,
      table: this.extractTable(query),
      parameters,
    });
  }

  logQueryError(error: string | Error, query: string, parameters?: any[]) {
    const context = this.getContext();
    this.logger.error(`[DB][ERROR] ${query}`, {
      ...context,
      table: this.extractTable(query),
      parameters,
      error,
    });
  }

  logQuerySlow(time: number, query: string, parameters?: any[]) {
    const context = this.getContext();
    const severity = time > ERROR_THRESHOLD ? 'error' : 'warn';
    const label = severity === 'error' ? 'VERY SLOW' : 'SLOW';

    this.logger[severity](`[DB][${label} > ${time}ms] ${query}`, {
      ...context,
      table: this.extractTable(query),
      parameters,
      duration: `${time}ms`,
      severity,
    });
  }

  logSchemaBuild(message: string) {
    this.logger.info(`[DB][SCHEMA] ${message}`);
  }

  logMigration(message: string) {
    this.logger.info(`[DB][MIGRATION] ${message}`);
  }

  log(level: 'log' | 'info' | 'warn', message: any) {
    this.logger.log({
      level,
      message: `[DB] ${message}`,
    });
  }
}
