import { FindManyOptions, Repository, SelectQueryBuilder } from 'typeorm';
import { InternalServerErrorException } from '@nestjs/common';
import { BaseResponseDto } from '../dto/base.response.dto';
import { PageResponseDto } from '../dto/pagination.response.dto';
import { buildTypeORMSortQuery } from '../util/buildTypeORMSortQuery';
import { buildTypeORMIncludeQuery } from '../util/buildTypeORMIncludeQuery';
import { buildTypeORMWhereClause } from '../util/buildTypeORMWhereClause';
import { getPaginationResponseDto } from '../util/getPaginationResponseDto';

export interface TypeORMCustomModelFindAllQuery {
  page: number | undefined;
  size: number | undefined;
  sort?: string;
  modelName?: string;
  include?: string;
  where?: { [key: string]: unknown };
  select?: string[];
  includeDisable?: boolean | string;
  includeSpaces?: boolean;
}

export interface ExtendedTypeORMCustomModelFindAllQuery
  extends TypeORMCustomModelFindAllQuery {
  search?: string;
}

interface CustomFindAllQuery {
  page?: number;
  size?: number;
  [key: string]: any;
}

interface FindAllQueryWithDefaults extends CustomFindAllQuery {
  page: number;
  size: number;
}

function getDefaultQueryOptions(
  query: Partial<TypeORMCustomModelFindAllQuery>,
): FindManyOptions & FindAllQueryWithDefaults {
  const { page, size, includeDisable, include, modelName, ...rest } = query;

  // Set default if undefined or null
  const returnPage = page ? Number(page) : 1;
  const returnSize = size ? Number(size) : 10;
  const returnIncludeDisable =
    includeDisable === true || includeDisable === 'true';

  // Return query with defaults and ensure modelName is passed through
  return {
    skip: (returnPage - 1) * returnSize,
    take: returnSize,
    where: {
      ...rest,
    },
    page: returnPage,
    size: returnSize,
    include: include || undefined,
    includeDisable: returnIncludeDisable,

    modelName: modelName || query.modelName, // Ensure modelName is passed through
  };
}

export interface TypeORMCustomModelFindAllQueryWithDefault
  extends TypeORMCustomModelFindAllQuery {
  page: number;
  size: number;
}

export type TypeORMCustomModelFindAllResponse = {
  baseResponseDto: BaseResponseDto;
  paginationResponseDto: PageResponseDto;
};

export function TypeORMCustomModel(repository: Repository<any>) {
  return Object.assign(repository, {
    findAll: async function (
      query: Partial<TypeORMCustomModelFindAllQuery>,
      customQueryBuilder?: SelectQueryBuilder<any>,
    ): Promise<TypeORMCustomModelFindAllResponse> {
      const {
        page = 1,
        size = 10,
        sort,
        modelName,
        include,
        where,
        select,
      } = getDefaultQueryOptions(query);

      if (!modelName) {
        throw new InternalServerErrorException(
          `[TypeORMCustomModel] Cannot findAll with unknown modelName`,
        );
      }

      const skip = (page - 1) * size;
      const order = buildTypeORMSortQuery(sort);
      const relations = buildTypeORMIncludeQuery(modelName, include);
      const whereClause = buildTypeORMWhereClause({ where });

      let data: any[] = [];
      let count = 0;

      if (customQueryBuilder) {
        const qb = customQueryBuilder.skip(skip).take(size);

        if (order) {
          Object.keys(order).forEach((key) => {
            qb.addOrderBy(key, order[key]);
          });
        }

        if (whereClause) {
          qb.andWhere(whereClause); // Use .andWhere instead of .where to avoid overwriting conditions
        }

        if (select) {
          const selectColumns = Array.isArray(select)
            ? select
            : Object.keys(select).map(
                (key) => `${customQueryBuilder.alias}.${key}`,
              );
          qb.select(selectColumns as string[]);
        }

        [data, count] = await qb.getManyAndCount();
      } else {
        [data, count] = await repository.findAndCount({
          where: whereClause,
          take: size,
          skip: skip,
          order: order,
          select: select,
          relations: relations,
        });
      }

      const paginationResponseDto = getPaginationResponseDto(count, page, size);
      const baseResponseDto: BaseResponseDto = {
        data,
        message: getResponseMessage(modelName, { where }),
      };

      return { baseResponseDto, paginationResponseDto };
    },
  });
}

function getResponseMessage(
  modelName: string,
  query?: {
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    where?: any;
  },
): string {
  if (!query) {
    return `Success get list ${modelName}`;
  }

  const { where } = query;
  if (modelName === 'user' && where && where?.community) {
    const {
      some: { communityId },
    } = where.community;
    if (typeof communityId === 'string') {
      return `Success get list ${modelName} belong to community`;
    }
  }

  return `Success get list ${modelName}`;
}
