WITH presence_logs AS (
  SELECT 
    d.space_device_uuid AS space_id,
    l.device_id,
    l.event_time,
    l.value,
    LAG(l.event_time) OVER (PARTITION BY l.device_id ORDER BY l.event_time) AS prev_time,
    LAG(l.value) OVER (PARTITION BY l.device_id ORDER BY l.event_time) AS prev_value 
  FROM device d
  JOIN "device-status-log" l ON d.uuid = l.device_id
  JOIN product p ON p.uuid = d.product_device_uuid
  WHERE l.code = 'presence_state'
    AND p.cat_name = 'hps'
),

-- Intervals when device was in 'presence' (between prev_time and event_time when value='none')
presence_intervals AS (
  SELECT 
    space_id,
    prev_time AS start_time,
    event_time AS end_time
  FROM presence_logs
  WHERE value = 'none' 
    AND prev_value = 'presence'
    AND prev_time IS NOT NULL
),

-- Split intervals across days
split_intervals AS (
  SELECT 
    space_id,
    generate_series(
      date_trunc('day', start_time),
      date_trunc('day', end_time),
      interval '1 day'
    )::date AS event_date,
    GREATEST(start_time, date_trunc('day', start_time)) AS interval_start,
    LEAST(end_time, date_trunc('day', end_time) + interval '1 day') AS interval_end
  FROM presence_intervals
),

-- Mark and group overlapping intervals per space per day
ordered_intervals AS (
  SELECT 
    space_id,
    event_date,
    interval_start,
    interval_end,
    LAG(interval_end) OVER (PARTITION BY space_id, event_date ORDER BY interval_start) AS prev_end
  FROM split_intervals
),

grouped_intervals AS (
  SELECT *,
    SUM(CASE 
          WHEN prev_end IS NULL OR interval_start > prev_end THEN 1 
          ELSE 0 
        END) OVER (PARTITION BY space_id, event_date ORDER BY interval_start) AS grp
  FROM ordered_intervals
),

-- Merge overlapping intervals per group
merged_intervals AS (
  SELECT 
    space_id,
    event_date,
    MIN(interval_start) AS merged_start,
    MAX(interval_end) AS merged_end
  FROM grouped_intervals
  GROUP BY space_id, event_date, grp
),

-- Sum durations of merged intervals
summed_intervals AS (
  SELECT
    space_id,
    event_date,
    SUM(EXTRACT(EPOCH FROM (merged_end - merged_start))) AS raw_occupied_seconds
  FROM merged_intervals
  GROUP BY space_id, event_date
  ),

final_data AS (
SELECT 
  space_id,
  event_date,
  LEAST(raw_occupied_seconds, 86400) AS occupied_seconds,
  ROUND(LEAST(raw_occupied_seconds, 86400) / 86400.0 * 100, 2) AS occupancy_percentage
FROM summed_intervals
ORDER BY space_id, event_date)


INSERT INTO public."space-daily-occupancy-duration" (
	space_uuid,	
	event_date,
	occupied_seconds,
	occupancy_percentage
)
select space_id,
       event_date,
       occupied_seconds,
       occupancy_percentage
FROM final_data
ON CONFLICT (space_uuid, event_date) DO UPDATE
SET
    occupancy_percentage = EXCLUDED.occupancy_percentage, 
   	occupied_seconds = EXCLUDED.occupied_seconds;
    
   
   