import { NestFactory } from '@nestjs/core';
import { AppModule } from './app.module';
import rateLimit from 'express-rate-limit';
import helmet from 'helmet';
import { setupSwaggerAuthentication } from '../libs/common/src/util/user-auth.swagger.utils';
import { ValidationPipe } from '@nestjs/common';
import { json, urlencoded } from 'body-parser';
import { SeederService } from '@app/common/seed/services/seeder.service';
import { HttpExceptionFilter } from './common/filters/http-exception/http-exception.filter';
import { Logger } from '@nestjs/common';
import { WINSTON_MODULE_NEST_PROVIDER } from 'nest-winston';
import { RequestContextMiddleware } from '@app/common/middleware/request-context.middleware';

async function bootstrap() {
  const app = await NestFactory.create(AppModule);

  app.enableCors();

  // Set the body parser limit to 1 MB
  app.use(json({ limit: '1mb' }));
  app.use(urlencoded({ limit: '1mb', extended: true }));
  app.useGlobalFilters(new HttpExceptionFilter());

  app.use(new RequestContextMiddleware().use);

  app.use(
    rateLimit({
      windowMs: 5 * 60 * 1000,
      max: 500,
    }),
  );

  app.use(
    helmet({
      contentSecurityPolicy: false,
    }),
  );

  setupSwaggerAuthentication(app);

  app.useGlobalPipes(
    new ValidationPipe({
      transform: true, // Auto-transform payloads to their DTO instances.
      transformOptions: {
        enableImplicitConversion: true, // Convert incoming payloads to their DTO instances if possible.
      },
    }),
  );

  const seederService = app.get(SeederService);
  const logger = app.get<Logger>(WINSTON_MODULE_NEST_PROVIDER);

  try {
    await seederService.seed();
    logger.log('Seeding complete!');
  } catch (error) {
    logger.error('Seeding failed!', error.stack || error);
  }

  logger.log('Starting auth at port ...', process.env.PORT || 4000);
  await app.listen(process.env.PORT || 4000);
}
bootstrap();
