import { SpaceProductAllocationRepository } from '@app/common/modules/space';
import { SpaceProductAllocationEntity } from '@app/common/modules/space/entities/space-product-allocation.entity';
import { SpaceEntity } from '@app/common/modules/space/entities/space.entity';
import { HttpException, HttpStatus, Injectable } from '@nestjs/common';

import { In, QueryRunner } from 'typeorm';

@Injectable()
export class SpaceProductAllocationService {
  constructor(
    private readonly spaceProductAllocationRepository: SpaceProductAllocationRepository,
  ) {}

  async createProductAllocations(
    space: SpaceEntity,
    allocationsData: { product: string; tag: string }[],
    queryRunner: QueryRunner,
  ): Promise<void> {
    try {
      if (!allocationsData.length) return;

      const productAllocations: SpaceProductAllocationEntity[] = [];

      for (const allocationData of allocationsData) {
        if (await this.isAllocationExist(queryRunner, allocationData, space))
          continue;

        const allocation = this.createNewAllocation(
          space,
          allocationData,
          queryRunner,
        );
        productAllocations.push(allocation);
      }
      if (productAllocations.length > 0) {
        await this.saveAllocations(productAllocations, queryRunner);
      }
    } catch (error) {
      throw this.handleError(
        error,
        'Failed to create space product allocations',
      );
    }
  }
  async unlinkModels(space: SpaceEntity, queryRunner: QueryRunner) {
    try {
      if (!space.productAllocations || space.productAllocations.length === 0)
        return;

      const allocationUuids = space.productAllocations.map(
        (allocation) => allocation.uuid,
      );

      await queryRunner.manager.update(
        SpaceProductAllocationEntity,
        { uuid: In(allocationUuids) },
        { inheritedFromModel: null },
      );
    } catch (error) {
      throw new HttpException(
        'Failed to unlink models',
        HttpStatus.INTERNAL_SERVER_ERROR,
      );
    }
  }

  private async isAllocationExist(
    queryRunner: QueryRunner,
    allocation: { product: string; tag: string },
    space: SpaceEntity,
  ): Promise<boolean> {
    const existingAllocations =
      await queryRunner.manager.findOne<SpaceProductAllocationEntity>(
        SpaceProductAllocationEntity,
        {
          where: {
            space: {
              uuid: space.uuid,
            },
            tag: {
              uuid: allocation.tag,
            },
            product: {
              uuid: allocation.product,
            },
          },
        },
      );

    return existingAllocations ? true : false;
  }

  async saveAllocations(
    allocations: SpaceProductAllocationEntity[],
    queryRunner?: QueryRunner,
  ) {
    queryRunner
      ? await queryRunner.manager.save(
          SpaceProductAllocationEntity,
          allocations,
        )
      : await this.spaceProductAllocationRepository.save(allocations);
  }

  async clearAllAllocations(spaceUuid: string, queryRunner: QueryRunner) {
    try {
      await queryRunner.manager.delete(SpaceProductAllocationEntity, {
        space: { uuid: spaceUuid },
      });
    } catch (error) {
      throw this.handleError(error, 'Failed to clear all allocations');
    }
  }

  private createNewAllocation(
    space: SpaceEntity,
    allocationData: { product: string; tag: string },
    queryRunner?: QueryRunner,
  ): SpaceProductAllocationEntity {
    return queryRunner
      ? queryRunner.manager.create(SpaceProductAllocationEntity, {
          space,
          product: { uuid: allocationData.product },
          tag: { uuid: allocationData.tag },
        })
      : this.spaceProductAllocationRepository.create({
          space,
          product: { uuid: allocationData.product },
          tag: { uuid: allocationData.tag },
        });
  }

  private handleError(error: any, message: string): HttpException {
    return new HttpException(
      error instanceof HttpException ? error.message : message,
      error instanceof HttpException
        ? error.getStatus()
        : HttpStatus.INTERNAL_SERVER_ERROR,
    );
  }
}
