import { BaseResponseDto } from '@app/common/dto/base.response.dto';
import { PageResponse } from '@app/common/dto/pagination.response.dto';
import { SuccessResponseDto } from '@app/common/dto/success.response.dto';
import {
  TypeORMCustomModel,
  TypeORMCustomModelFindAllQuery,
} from '@app/common/models/typeOrmCustom.model';
import { SubspaceDto } from '@app/common/modules/space/dtos';
import { HttpException, HttpStatus, Injectable } from '@nestjs/common';
import { In, Not, QueryRunner } from 'typeorm';
import {
  AddSubspaceDto,
  GetSpaceParam,
  GetSubSpaceParam,
  ModifySubspaceDto,
} from '../../dtos';

import { SubspaceModelEntity } from '@app/common/modules/space-model';
import { SpaceEntity } from '@app/common/modules/space/entities/space.entity';
import { SubspaceProductAllocationEntity } from '@app/common/modules/space/entities/subspace/subspace-product-allocation.entity';
import { SubspaceEntity } from '@app/common/modules/space/entities/subspace/subspace.entity';
import { SubspaceRepository } from '@app/common/modules/space/repositories/subspace.repository';
import {
  AllocationsOwnerType,
  CreateAllocationsDto,
} from 'src/space/dtos/create-allocations.dto';
import { TagService as NewTagService } from 'src/tags/services/tags.service';
import { ValidationService } from '../space-validation.service';
import { SubspaceDeviceService } from './subspace-device.service';
import { SubspaceProductAllocationService } from './subspace-product-allocation.service';

@Injectable()
export class SubSpaceService {
  constructor(
    private readonly subspaceRepository: SubspaceRepository,
    private readonly validationService: ValidationService,
    private readonly newTagService: NewTagService,
    public readonly deviceService: SubspaceDeviceService,
    private readonly subspaceProductAllocationService: SubspaceProductAllocationService,
  ) {}

  async createSubspaces(
    subspaceData: Array<{
      subspaceName: string;
      space: SpaceEntity;
      subSpaceModel?: SubspaceModelEntity;
    }>,
    queryRunner: QueryRunner,
  ): Promise<SubspaceEntity[]> {
    try {
      const subspaceNames = subspaceData.map((data) => data.subspaceName);

      await this.checkExistingNamesInSpace(
        subspaceNames,
        subspaceData[0].space,
      );

      const subspaces = subspaceData.map((data) =>
        queryRunner.manager.create(SubspaceEntity, data),
      );
      return queryRunner.manager.save(subspaces);
    } catch (error) {
      throw new HttpException(
        `An unexpected error occurred while creating subspaces. ${error}`,
        HttpStatus.INTERNAL_SERVER_ERROR,
      );
    }
  }

  async createSubSpaceFromModel(
    subspaceModels: SubspaceModelEntity[],
    spaces: SpaceEntity[],
    queryRunner: QueryRunner,
  ): Promise<void> {
    if (!subspaceModels?.length || !spaces?.length) return;

    const subspaceData = [];

    for (const space of spaces) {
      for (const subSpaceModel of subspaceModels) {
        subspaceData.push({
          subspaceName: subSpaceModel.subspaceName,
          space,
          subSpaceModel,
        });
      }
    }
    await this.createSubspaces(subspaceData, queryRunner);
  }

  async createSubspacesFromDto(
    addSubspaceDtos: AddSubspaceDto[],
    space: SpaceEntity,
    queryRunner: QueryRunner,
    projectUuid?: string,
  ): Promise<SubspaceEntity[]> {
    try {
      const createdSubspaces = await this.createSubspaces(
        addSubspaceDtos.map((dto) => ({
          subspaceName: dto.subspaceName,
          space,
        })),
        queryRunner,
      );
      await Promise.all(
        addSubspaceDtos.map(async ({ tags }, index) => {
          // map the dto to the corresponding subspace
          const subspace = createdSubspaces[index];
          await this.createAllocations({
            projectUuid,
            queryRunner,
            tags,
            type: AllocationsOwnerType.SUBSPACE,
            subspace,
          });
        }),
      );
      return createdSubspaces;
    } catch (error) {
      if (error instanceof HttpException) {
        throw error;
      }

      throw new HttpException(
        'Failed to save subspaces due to an unexpected error.',
        HttpStatus.INTERNAL_SERVER_ERROR,
      );
    }
  }

  async createSubspace(
    addSubspaceDto: AddSubspaceDto,
    params: GetSpaceParam,
  ): Promise<BaseResponseDto> {
    const space =
      await this.validationService.validateSpaceWithinCommunityAndProject(
        params.communityUuid,
        params.projectUuid,
        params.spaceUuid,
      );

    try {
      await this.checkExistingNamesInSpace(
        [addSubspaceDto.subspaceName],
        space,
      );
      const newSubspace = this.subspaceRepository.create({
        ...addSubspaceDto,
        space,
      });

      await this.subspaceRepository.save(newSubspace);

      return new SuccessResponseDto({
        statusCode: HttpStatus.CREATED,
        data: newSubspace,
        message: 'Subspace created successfully',
      });
    } catch (error) {
      throw new HttpException(error.message, HttpStatus.INTERNAL_SERVER_ERROR);
    }
  }

  async list(
    params: GetSpaceParam,
    pageable: Partial<TypeORMCustomModelFindAllQuery>,
  ): Promise<BaseResponseDto> {
    const { communityUuid, spaceUuid, projectUuid } = params;
    await this.validationService.checkCommunityAndProjectSpaceExistence(
      communityUuid,
      projectUuid,
      spaceUuid,
    );

    try {
      pageable.modelName = 'subspace';
      pageable.where = { space: { uuid: spaceUuid }, disabled: false };
      const customModel = TypeORMCustomModel(this.subspaceRepository);

      const { baseResponseDto, paginationResponseDto } =
        await customModel.findAll(pageable);
      return new PageResponse<SubspaceDto>(
        baseResponseDto,
        paginationResponseDto,
      );
    } catch (error) {
      throw new HttpException(error.message, HttpStatus.INTERNAL_SERVER_ERROR);
    }
  }

  async updateSubSpace(
    params: GetSubSpaceParam,
    updateSubSpaceDto: AddSubspaceDto,
  ): Promise<BaseResponseDto> {
    const { spaceUuid, communityUuid, subSpaceUuid, projectUuid } = params;
    await this.validationService.checkCommunityAndProjectSpaceExistence(
      communityUuid,
      projectUuid,
      spaceUuid,
    );
    const subSpace = await this.subspaceRepository.findOne({
      where: { uuid: subSpaceUuid, disabled: false },
    });

    if (!subSpace) {
      throw new HttpException(
        `SubSpace with ID ${subSpaceUuid} not found`,
        HttpStatus.NOT_FOUND,
      );
    }

    try {
      Object.assign(subSpace, updateSubSpaceDto);

      await this.subspaceRepository.save(subSpace);

      return new SuccessResponseDto({
        message: `Subspace with ID ${subSpaceUuid} successfully updated`,
        data: subSpace,
        statusCode: HttpStatus.OK,
      });
    } catch (error) {
      if (error instanceof HttpException) {
        throw error;
      }
      throw new HttpException(
        'An error occurred while updating the subspace',
        HttpStatus.INTERNAL_SERVER_ERROR,
      );
    }
  }

  async delete(params: GetSubSpaceParam): Promise<BaseResponseDto> {
    const { spaceUuid, communityUuid, subSpaceUuid, projectUuid } = params;
    await this.validationService.checkCommunityAndProjectSpaceExistence(
      communityUuid,
      projectUuid,
      spaceUuid,
    );

    const subspace = await this.subspaceRepository.findOne({
      where: { uuid: subSpaceUuid, disabled: false },
    });

    if (!subspace) {
      throw new HttpException(
        `Subspace with ID ${subSpaceUuid} not found`,
        HttpStatus.NOT_FOUND,
      );
    }

    try {
      await this.subspaceRepository.remove(subspace);

      return new SuccessResponseDto({
        message: `Subspace with ID ${subSpaceUuid} successfully deleted`,
        statusCode: HttpStatus.OK,
      });
    } catch (error) {
      throw new HttpException(
        'An error occurred while deleting the subspace',
        HttpStatus.INTERNAL_SERVER_ERROR,
      );
    }
  }

  /*  async deleteSubspaces(
    deleteDtos: DeleteSubspaceDto[],
    queryRunner: QueryRunner,
  ) {
    /*   const deleteResults: { uuid: string }[] = [];

    for (const dto of deleteDtos) {
      const subspace = await this.findOne(dto.subspaceUuid);

      await queryRunner.manager.update(
        this.subspaceRepository.target,
        { uuid: dto.subspaceUuid },
        { disabled: true },
      );

      if (subspace.tags?.length) {
        const modifyTagDtos = subspace.tags.map((tag) => ({
          uuid: tag.uuid,
          action: ModifyAction.DELETE,
        }));
        await this.tagService.modifyTags(
          modifyTagDtos,
          queryRunner,
          null,
          subspace,
        );
      }

      if (subspace.devices)
        await this.deviceService.deleteSubspaceDevices(
          subspace.devices,
          queryRunner,
        );

      deleteResults.push({ uuid: dto.subspaceUuid });
    }

    return deleteResults;
  } */

  async updateSubspaceInSpace(
    subspaceDtos: ModifySubspaceDto[],
    queryRunner: QueryRunner,
    space: SpaceEntity,
    projectUuid: string,
  ) {
    try {
      // disable subspaces that are not in the provided list & delte their allocations
      await queryRunner.manager.update(
        SubspaceEntity,
        {
          uuid: Not(
            In(subspaceDtos.filter(({ uuid }) => uuid).map(({ uuid }) => uuid)),
          ),
          space: { uuid: space.uuid },
        },
        {
          disabled: true,
        },
      );

      await queryRunner.manager.delete(SubspaceProductAllocationEntity, {
        subspace: { uuid: Not(In(subspaceDtos.map((dto) => dto.uuid))) },
      });

      // create or update subspaces provided in the list
      const newSubspaces = this.subspaceRepository.create(
        subspaceDtos.filter((dto) => !dto.uuid),
      );

      const updatedSubspaces: SubspaceEntity[] = await queryRunner.manager.save(
        SubspaceEntity,
        [...newSubspaces, ...subspaceDtos.filter((dto) => dto.uuid)].map(
          (subspace) => ({ ...subspace, space }),
        ),
      );

      // create or update allocations for the subspaces
      if (updatedSubspaces.length > 0) {
        await this.subspaceProductAllocationService.updateSubspaceProductAllocationsV2(
          subspaceDtos.map((dto) => {
            if (!dto.uuid) {
              dto.uuid = updatedSubspaces.find(
                (subspace) => subspace.subspaceName === dto.subspaceName,
              )?.uuid;
            }
            return {
              tags: dto.tags || [],
              uuid: dto.uuid,
            };
          }),
          projectUuid,
          queryRunner,
        );
      }
    } catch (error) {
      throw new HttpException(
        `An error occurred while modifying subspaces: ${error.message}`,
        HttpStatus.INTERNAL_SERVER_ERROR,
      );
    }
  }

  async unlinkModels(
    subspaces: SubspaceEntity[],
    queryRunner: QueryRunner,
  ): Promise<void> {
    if (!subspaces || subspaces.length === 0) return;

    try {
      const subspaceUuids = subspaces.map((subSpace) => subSpace.uuid);

      const allocations: SubspaceProductAllocationEntity[] = subspaces.flatMap(
        (subspace) => subspace.productAllocations || [],
      );

      if (allocations.length > 0) {
        await this.subspaceProductAllocationService.unlinkModels(
          allocations,
          queryRunner,
        );
      }

      await queryRunner.manager.update(
        SubspaceEntity,
        { uuid: In(subspaceUuids) },
        { subSpaceModel: null },
      );
    } catch (error) {
      throw new HttpException(
        `Failed to unlink subspace models: ${error instanceof Error ? error.message : 'Unknown error'}`,
        HttpStatus.INTERNAL_SERVER_ERROR,
      );
    }
  }

  async getOne(params: GetSubSpaceParam): Promise<BaseResponseDto> {
    await this.validationService.checkCommunityAndProjectSpaceExistence(
      params.communityUuid,
      params.projectUuid,
      params.spaceUuid,
    );
    const subspace = await this.findOne(params.subSpaceUuid);
    return new SuccessResponseDto({
      message: `Successfully retrieved subspace`,
      data: subspace,
    });
  }

  private async findOne(subspaceUuid: string): Promise<SubspaceEntity> {
    const subspace = await this.subspaceRepository.findOne({
      where: { uuid: subspaceUuid, disabled: false },
      relations: ['space', 'devices'],
    });
    if (!subspace) {
      throw new HttpException(
        `SubspaceModel with UUID ${subspaceUuid} not found.`,
        HttpStatus.NOT_FOUND,
      );
    }
    return subspace;
  }

  private async checkExistingNamesInSpace(
    names: string[],
    space: SpaceEntity,
  ): Promise<void> {
    const existingNames = await this.subspaceRepository.find({
      select: ['subspaceName'],
      where: {
        subspaceName: In(names),
        space: {
          uuid: space.uuid,
        },
        disabled: false,
      },
    });

    if (existingNames.length > 0) {
      const existingNamesList = existingNames
        .map((e) => e.subspaceName)
        .join(', ');
      throw new HttpException(
        `Subspace names already exist in the space: ${existingNamesList}`,
        HttpStatus.BAD_REQUEST,
      );
    }
  }

  async clearSubspaces(subspaceUuids: string[], queryRunner: QueryRunner) {
    try {
      await queryRunner.manager.update(
        SubspaceEntity,
        { uuid: In(subspaceUuids) },
        { disabled: true },
      );

      await this.subspaceProductAllocationService.clearAllAllocations(
        subspaceUuids,
        queryRunner,
      );
    } catch (error) {
      throw new HttpException(
        error instanceof HttpException
          ? error.message
          : 'An unexpected error occurred while clearing subspaces',
        error instanceof HttpException
          ? error.getStatus()
          : HttpStatus.INTERNAL_SERVER_ERROR,
      );
    }
  }

  async createAllocations(dto: CreateAllocationsDto): Promise<void> {
    const { projectUuid, queryRunner, tags, type } = dto;

    const allocationsData = await this.newTagService.processTags(
      tags,
      projectUuid,
      queryRunner,
    );

    // Create a mapping of created tags by UUID and name for quick lookup
    const createdTagsByUUID = new Map(allocationsData.map((t) => [t.uuid, t]));
    const createdTagsByName = new Map(allocationsData.map((t) => [t.name, t]));

    // Create the product-tag mapping based on the processed tags
    const productTagMapping = tags.map(({ uuid, name, productUuid }) => {
      const inputTag = uuid
        ? createdTagsByUUID.get(uuid)
        : createdTagsByName.get(name);
      return {
        tag: inputTag?.uuid,
        product: productUuid,
      };
    });

    switch (type) {
      case AllocationsOwnerType.SUBSPACE: {
        await this.subspaceProductAllocationService.createProductAllocations(
          dto.subspace,
          productTagMapping,
          queryRunner,
        );
        break;
      }
    }
  }
}
