import { HttpException, HttpStatus, Injectable } from '@nestjs/common';
import { UserDevicePermissionAddDto } from '../dtos/user-device-permission.add.dto';
import { UserDevicePermissionEditDto } from '../dtos/user-device-permission.edit.dto';
import { DeviceUserPermissionRepository } from '@app/common/modules/device/repositories';
import { PermissionTypeRepository } from '@app/common/modules/permission/repositories';
import { CommonErrorCodes } from '@app/common/constants/error-codes.enum';

@Injectable()
export class UserDevicePermissionService {
  constructor(
    private readonly deviceUserPermissionRepository: DeviceUserPermissionRepository,
    private readonly permissionTypeRepository: PermissionTypeRepository,
  ) {}

  async addUserPermission(userDevicePermissionDto: UserDevicePermissionAddDto) {
    try {
      const permissionType = await this.getPermissionType(
        userDevicePermissionDto.permissionType,
      );
      return await this.deviceUserPermissionRepository.save({
        userUuid: userDevicePermissionDto.userUuid,
        deviceUuid: userDevicePermissionDto.deviceUuid,
        permissionType: {
          uuid: permissionType.uuid,
        },
      });
    } catch (error) {
      if (error.code === CommonErrorCodes.DUPLICATE_ENTITY) {
        throw new HttpException(
          'This User already belongs to this device',
          HttpStatus.BAD_REQUEST,
        );
      }
      throw new HttpException(
        error.message || 'Internal Server Error',
        HttpStatus.INTERNAL_SERVER_ERROR,
      );
    }
  }

  async editUserPermission(
    devicePermissionUuid: string,
    userDevicePermissionEditDto: UserDevicePermissionEditDto,
  ) {
    try {
      const permissionType = await this.getPermissionType(
        userDevicePermissionEditDto.permissionType,
      );
      return await this.deviceUserPermissionRepository.update(
        { uuid: devicePermissionUuid },
        {
          permissionType: {
            uuid: permissionType.uuid,
          },
        },
      );
    } catch (error) {
      if (error.code === CommonErrorCodes.DUPLICATE_ENTITY) {
        throw new HttpException(
          'This User already belongs to this device',
          HttpStatus.BAD_REQUEST,
        );
      }
      throw new HttpException(
        error.message || 'Internal Server Error',
        HttpStatus.INTERNAL_SERVER_ERROR,
      );
    }
  }

  async fetchUserPermission(deviceUuid: string) {
    const devicePermissions = await this.deviceUserPermissionRepository.find({
      where: {
        deviceUuid: deviceUuid,
      },
      relations: ['permissionType', 'user'],
    });
    return devicePermissions.map((permission) => {
      return {
        uuid: permission.uuid,
        deviceUuid: permission.deviceUuid,
        firstName: permission.user.firstName,
        lastName: permission.user.lastName,
        email: permission.user.email,
        permissionType: permission.permissionType.type,
      };
    });
  }
  private async getPermissionType(permissionType: string) {
    return await this.permissionTypeRepository.findOne({
      where: {
        type: permissionType,
      },
    });
  }
  async deleteDevicePermission(devicePermissionUuid: string) {
    try {
      return await this.deviceUserPermissionRepository.delete({
        uuid: devicePermissionUuid,
      });
    } catch (error) {
      throw new HttpException(
        error.message || 'Internal Server Error',
        HttpStatus.INTERNAL_SERVER_ERROR,
      );
    }
  }
}
